#!/usr/bin/python
# coding=utf-8

from __future__ import division, print_function, unicode_literals

__all__ = ["__version__", "__author__", "__author_email__"]

__version__ = "0.5"

__author__ = 'Klaus Greff'
__author_email__ = 'qwlouse@gmail.com',

__url__ = "https://github.com/Qwlouse/sacred"
